from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys
import os
filepath = os.path.abspath(__file__)
prj_dir = os.path.abspath(os.path.join(os.path.dirname(filepath), "../"))
lib_dir = os.path.abspath(os.path.join(prj_dir, "lib"))
sys.path.append(prj_dir)
sys.path.append(lib_dir)

import torch
import numpy as np
import cv2
import vot

from lib.test.evaluation import Tracker
from VOT22.SBT.SBTv1 import build_SBTv1_base_model
from VOT22.SBT.tracker_SBTv1 import Tracker_SBTv1
from ltr.dataset.depth_utils import get_rgbd_frame


def bbox_clip(x1, y1, x2, y2, boundary, min_sz=10):
    '''boundary (H,W)'''
    x1_new = max(0, min(x1, boundary[1] - min_sz))
    y1_new = max(0, min(y1, boundary[0] - min_sz))
    x2_new = max(min_sz, min(x2, boundary[1]))
    y2_new = max(min_sz, min(y2, boundary[0]))
    return x1_new, y1_new, x2_new, y2_new

def read_image(image_file, dtype='color'):
#     print(image_file)
    if isinstance(image_file, list):
        img = get_rgbd_frame(image_file[0], image_file[1], dtype=dtype, depth_clip=True)
    else:
        if dtype == 'color':
            img = get_rgbd_frame(image_file, None, dtype=dtype, depth_clip=False)
        else:
            img = get_rgbd_frame(None, image_file, dtype=dtype, depth_clip=False)
    return img


class MotionEstimator(object):
    """ Statistic of the track """
    def __init__(self):
        self.pos = None
        self.vel = None
        self.conf = 0
        self.vel_his = []  # velocity history
        self.vel_his_conf = []  # velocity history

    def initialize(self, box):
        self.pos = box[:2] + box[2:]/2
        self.wh = box[2:]
        self.vel_his = []  # velocity history
        self.vel_his_conf = []  # velocity history

    def update_vel_his(self, v, conf):
        self.vel_his.append(v)
        self.vel_his_conf.append(conf)
        if len(self.vel_his) > 20:
            self.vel_his.pop(0)
            self.vel_his_conf.pop(0)
            
        vel_his = np.stack(self.vel_his)
        vel_conf = np.stack(self.vel_his_conf)
        vel_conf_norm = vel_conf/vel_conf.sum()

        vel_mean = vel_conf_norm.dot(vel_his)
        vel_var = vel_his.var(0)
        return vel_mean, vel_var

    def cosine_pdf(self, x, x_, size):
        """
        Args:
            x: Tracker预测位置
            x_: 运动模型预测位置
            size: 目标尺寸
        """
        PI = 3.1415
        phase = min(np.linalg.norm(x-x_)/size, 1)*PI/2
        return np.cos(phase)

    def update(self, box, conf, th=0.3):
        pos = box[:2] + box[2:]/2
        if conf > th:
            self.wh = box[2:]

        _v = pos - self.pos  # 瞬时速度
        if self.vel is None:
            self.vel = _v  # 初始化estimator的速度
        vel_mean, vel_var = self.update_vel_his(_v, conf)  # 目标运动速度的统计量

        self.vel = vel_mean
        self.pos = pos
        self.conf = conf
        return box, self.conf

    def rescore(self, box, conf):
        pos = box[:2] + box[2:]/2

        _v = pos - self.pos  # 瞬时速度
        if self.vel is None:
            self.vel = _v  # 初始化estimator的速度

        motion_conf = self.cosine_pdf(pos, self.pos + self.vel, size=np.sqrt(self.wh.prod()))  # 当前目标运动速度的可信度
        print(motion_conf)
        conf = conf * (conf ** 2 + (1-conf**2) * motion_conf)
        return conf


def IOU(rec1, rec2):
    left_column_max  = max(rec1[0],rec2[0])
    right_column_min = min(rec1[0]+rec1[2],rec2[0] + rec2[2])
    up_row_max       = max(rec1[1],rec2[1])
    down_row_min     = min(rec1[1] + rec1[3],rec2[1] + rec2[3])
    #两矩形无相交区域的情况
    if left_column_max>=right_column_min or down_row_min<=up_row_max:
        return 0
    # 两矩形有相交区域的情况
    else:
        S1 = rec1[2]*rec1[3]
        S2 = rec2[2]*rec2[3]
        S_cross = (down_row_min-up_row_max)*(right_column_min-left_column_max)
        return S_cross/(S1+S2-S_cross)


class SBT_REF(object):
    """SBT Tracker"""
    def __init__(self, net_path=None, refine_path=None, ref_tracker='stark_ref', ref_param='baseline', use_new_box=True):
        """use_new_box: whether to use the refined box as the new state"""
        self.use_new_box = use_new_box
        # create base tracker
        
        net = build_SBTv1_base_model()
        net.load_state_dict(torch.load(net_path)['net'])
        net.cuda()
        self.tracker = Tracker_SBTv1(name='SBTv1_520', net=net, window_penalty=0.49, window_size=32, exemplar_size=128, instance_size=256, refine_path=refine_path)
        
        # create refinement module
        ref_info = Tracker(ref_tracker, ref_param, "lasot", None)
        ref_params = ref_info.get_parameters()
        ref_params.visualization, ref_params.debug = False, False
        self.ref = ref_info.create_tracker(ref_params)

        self.me = MotionEstimator()
        self.cfg = {}
        self.aux_conf_max = 0.8
        self.dimp_th = 0.999  # threshold for switch to dimp
        self.aux_conf_min = -0.0001
        self.cfg['aux_conf_max'] = self.aux_conf_max
        self.cfg['aux_conf_min'] = self.aux_conf_min

    def initialize(self, imgfile, box, aux_tracker):
        """box: list"""
        # init on the 1st frame
        img = read_image(imgfile, dtype='colormap')  # Right
        self.H, self.W, _ = img.shape
        init_info = {'init_bbox': box}
        
        # init base tracker
        _ = self.tracker.initialize(img, init_info)
        
        self.ref.initialize(img, init_info)
        self.aux_tracker = aux_tracker
        self.me.initialize(np.array(box))

    def track(self, imgpath):
        img = read_image(imgpath, dtype='colormap')
        # track with the base tracker
        pred_bbox, _, conf_score = self.tracker.track(img)

        
        aux_outputs = self.aux_tracker.track(img)
        aux_pred_bbox = aux_outputs['target_bbox']

        _conf_score = self.me.rescore(np.array(pred_bbox), conf_score)  # depress teleport  # TODO: check the rescore preocess

        update_flag = False

        if _conf_score < self.dimp_th:  
            pred_bbox = aux_pred_bbox
            self.aux_conf_max = max(self.aux_conf_max * 0.99, self.aux_conf_min)
            conf_score = min(aux_outputs['conf'], self.aux_conf_max)
        else:
            self.aux_conf_max = self.cfg['aux_conf_max']
        pred_bbox = [float(num) for num in pred_bbox]
        
        # refine with the refinement module
        outputs_ref = self.ref.track(img, pred_bbox, update_flag)
        pred_bbox_ref = outputs_ref["target_bbox"]

        pred_bbox_ref, _ = self.me.update(np.array(pred_bbox), conf_score)

        if self.use_new_box:
            self.tracker.center_pos = np.array([pred_bbox_ref[0]+pred_bbox_ref[2]/2, pred_bbox_ref[1]+pred_bbox_ref[3]/2])
            self.tracker.size = np.array([pred_bbox_ref[2], pred_bbox_ref[3]])

        self.aux_tracker.update_state(pred_bbox_ref)

        return pred_bbox_ref, conf_score, _conf_score


class DiMP():
    def __init__(self, img, init_box):
        # create tracker
        from pytracking.evaluation import Tracker

        tracker_info = Tracker('dimp', 'DeT_DiMP50_Mean', None)
        params = tracker_info.get_parameters()
        params.visualization = False
        params.debug = False
        params.visdom_info = {'use_visdom': False, 'server': '127.0.0.1', 'port': 8097}
        self.tracker = tracker_info.tracker_class(params)

        self.im_H, self.im_W, _ = img.shape
#         cx, cy, w, h = self.get_axis_aligned_bbox(np.array(init_box))
#         gt_bbox_ = [cx - (w - 1) / 2, cy - (h - 1) / 2, w, h]
#         '''Initialize'''
#         gt_bbox_np = np.array(gt_bbox_)
#         gt_bbox_torch = torch.from_numpy(gt_bbox_np.astype(np.float32))
        init_info = {}
        init_info['init_bbox'] = init_box
        self.tracker.initialize(img, init_info)

    def track(self, img):
        return self.tracker.track(img)

    @staticmethod
    def get_axis_aligned_bbox(region):
        """ convert region to (cx, cy, w, h) that represent by axis aligned box
        """
        nv = region.size
        if nv == 8:
            cx = np.mean(region[0::2])
            cy = np.mean(region[1::2])
            x1 = min(region[0::2])
            x2 = max(region[0::2])
            y1 = min(region[1::2])
            y2 = max(region[1::2])
            A1 = np.linalg.norm(region[0:2] - region[2:4]) * \
                 np.linalg.norm(region[2:4] - region[4:6])
            A2 = (x2 - x1) * (y2 - y1)
            s = np.sqrt(A1 / A2)
            w = s * (x2 - x1) + 1
            h = s * (y2 - y1) + 1
        else:
            x = region[0]
            y = region[1]
            w = region[2]
            h = region[3]
            cx = x + w / 2
            cy = y + h / 2
        return cx, cy, w, h

    @staticmethod
    def bbox_clip(x1, y1, x2, y2, boundary, min_sz=10):
        x1_new = max(0, min(x1, boundary[1] - min_sz))
        y1_new = max(0, min(y1, boundary[0] - min_sz))
        x2_new = max(min_sz, min(x2, boundary[1]))
        y2_new = max(min_sz, min(y2, boundary[0]))
        return x1_new, y1_new, x2_new, y2_new

    def update_state(self, pred_bbox):

        x1, y1, w, h = pred_bbox.tolist()
        '''add boundary and min size limit'''
        x1, y1, x2, y2 = self.bbox_clip(x1, y1, x1 + w, y1 + h, (self.im_H, self.im_W))
        w = x2 - x1
        h = y2 - y1
        new_pos = torch.from_numpy(np.array([y1 + h / 2, x1 + w / 2]).astype(np.float32))
        new_target_sz = torch.from_numpy(np.array([h, w]).astype(np.float32))
        new_scale = torch.sqrt(new_target_sz.prod() / self.tracker.base_target_sz.prod())
        ##### update
        self.tracker.pos = new_pos.clone()
        self.tracker.target_sz = new_target_sz
        self.tracker.target_scale = new_scale


def run_vot_exp(net_path, refine_path, use_new_box, save_dir, vis=False):
    torch.set_num_threads(1)
    dir_name = 'SBT_DeT_depth'
    save_root = os.path.join(save_dir, dir_name)
    tracker = SBT_REF(net_path=net_path, refine_path=refine_path, use_new_box=use_new_box)
    
    handle = vot.VOT("rectangle", "rgbd")
    selection = handle.region()
    imagefile = handle.frame()
    init_box = [selection.x, selection.y, selection.width, selection.height]
    
    if not imagefile:
        sys.exit(0)
    if vis:
        '''for vis'''
        seq_name = imagefile[0].split('/')[-3]
        save_dir = os.path.join(save_root, seq_name)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
    
    img = read_image(imagefile[1], dtype='colormap')
    dimp = DiMP(img, init_box)
    
    tracker.initialize(imagefile, init_box, dimp)

    cnt = 0
    while True:
        imagefile = handle.frame()
        if not imagefile:
            break
        b1, conf, _conf = tracker.track(imagefile)
        x1, y1, w, h = b1
        handle.report(vot.Rectangle(x1, y1, w, h), conf)
        if vis and (cnt % 5) == 0:
            '''Visualization'''
            # original image
            image = cv2.cvtColor(cv2.imread(imagefile[0]), cv2.COLOR_BGR2RGB)  # rgbs
            image_ori = image[:, :, ::-1].copy()  # RGB --> BGR
            image_name = imagefile[0].split('/')[-1]
            # tracker box
            image_b = image_ori.copy()
            cv2.rectangle(image_b, (int(b1[0]), int(b1[1])),
                          (int(b1[0] + b1[2]), int(b1[1] + b1[3])), (0, 0, 255), 2)
            cv2.putText(image_b, "{:.4f}".format(conf), (100, 100), cv2.FONT_HERSHEY_SIMPLEX, 1.2, (255, 255, 0), 2)
            cv2.putText(image_b, "{:.4f}".format(_conf), (100, 300), cv2.FONT_HERSHEY_SIMPLEX, 1.2, (0, 255, 0), 2)
            image_b_name = image_name
            save_path = os.path.join(save_dir, image_b_name)
            cv2.imwrite(save_path, image_b)
        cnt += 1


if __name__ == '__main__':
    save_dir = os.path.abspath(os.path.join(prj_dir, '../analysis/vis'))
    net_path = prj_dir+'/checkpoints/SBT/SBTv1_base_ep0533.pth.tar'
    refine_path = None
    run_vot_exp(net_path=net_path, refine_path=refine_path,
                use_new_box=True, save_dir=save_dir, vis=False)
